from joatmon.core import CoreException


class HZMQTTException(CoreException):
    pass


class MQTTException(CoreException):
    pass


class CodecException(CoreException):
    pass


class NoDataException(CoreException):
    pass
